//------------------------------------------------------------------------------
// ADC
//------------------------------------------------------------------------------
template<
    class Control,
    class Reference,
    class Prescaler,
    template<const basic_pin_id&> class ChannelLookup,
    class Resolution
> struct basic_adc {
    
    //--------------------------------------------------------------------------
    using reference      = Reference;
    using prescaler      = Prescaler;
    using resolution     = Resolution;

    //--------------------------------------------------------------------------
    template<class R>
    inline static void set_resolution() {
        Control::set_resolution::apply(R::value);
    }

    //--------------------------------------------------------------------------
    template<const basic_pin_id& Pin>
    inline static void select_channel() {
        Control::select_channel::apply(ChannelLookup<Pin>::value);
    }

    //--------------------------------------------------------------------------
    inline static bool conversion_running() {
        return Control::read_start_conversion::apply();
    }

    //--------------------------------------------------------------------------
    template<const typename reference::id_t& R>
    inline static void set_reference() {
        Control::set_reference::apply(*R);
    }

    //--------------------------------------------------------------------------
    template<const typename prescaler::id_t& PS>
    inline static void set_prescaler() {
        Control::set_prescaler::apply(*PS);
    }

    //--------------------------------------------------------------------------
    inline static void enable() {
        Control::enable::apply(1);
    }

    inline static void disable() {
        Control::enable::apply(0);
    }

    //--------------------------------------------------------------------------
    template<class Res>
    inline static typename Res::type read() {
        return Res::read();
    }

    //--------------------------------------------------------------------------
    inline static void start() {
        Control::start::apply();
    }
};
