//------------------------------------------------------------------------------
// BIT MASK
//------------------------------------------------------------------------------
namespace bit {
    enum class operation { set, clr };

    template<count_t Bit> struct set {
        constexpr static operation op = operation::set;
        constexpr static count_t bit = Bit;
    };

    template<count_t Bit> struct clr {
        constexpr static operation op = operation::clr;
        constexpr static count_t bit = Bit;
    };

    template<const detail::reg::id_t& Register, class Op>
    constexpr inline void _do_apply_op() {
        if (Op::op == operation::set) {
            reg::set<Register, Op::bit>();
        } else {
            reg::clr<Register, Op::bit>();
        }
    }

    template<const detail::reg::id_t& Register, class Head, class... Tail>
    struct _apply_write {
        inline _apply_write() {
            _do_apply_op<Register,Head>();
            _apply_write<Register, Tail...> {};
        }
    };

    template<const detail::reg::id_t& Register, class Head>
    struct _apply_write<Register, Head> {
        inline _apply_write() {
            _do_apply_op<Register,Head>();
        }
    };

    template<const detail::reg::id_t& Register, class... Op>
    struct write {
        inline static void apply() { _apply_write<Register, Op...> {}; };
    };

//------------------------------------------------------------------------------
    template<count_t Head, count_t... Tail>
    struct mask {
        constexpr static mc::word_t value = (1 << Head) | mask<Tail...>::value;
    };

    template<count_t Head>
    struct mask<Head> {
        constexpr static mc::word_t value = (1 << Head);
    };

    template<const detail::reg::id_t& Register, count_t... Ps>
    struct read {
        inline static mc::word_t apply() {
            return detail::reg::read<Register>() & mask<Ps...>::value;
        }
    };

    template<const detail::reg::id_t& Register, count_t... Ps>
    struct put {
        inline static void apply() {
            detail::reg::write<Register>(mask<Ps...>::value);
        }
    };

//------------------------------------------------------------------------------
} // namespace mc::bit
