//------------------------------------------------------------------------------
// BASIC IRQ SET
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
DEFINE_VALUE_ENUM(basic_irq_id,         count_t)
DEFINE_VALUE_ENUM(basic_irq_trigger_id, count_t)

using isr_t = void(*)();

struct isr_table {
    static isr_t value[];
};

//------------------------------------------------------------------------------
template<
    class Control,
    count_t Count
> struct basic_irq {

    //--------------------------------------------------------------------------
    using id_t  = basic_irq_id;
    using isr_t = mc::detail::isr_t;

    constexpr static count_t count = Count;

    //--------------------------------------------------------------------------
    struct trigger {
        using id_t = basic_irq_trigger_id;

        constexpr static const id_t low     = 0;
        constexpr static const id_t any     = 1;
        constexpr static const id_t falling = 2;
        constexpr static const id_t rising  = 3;
    };

    //--------------------------------------------------------------------------
    template<const id_t& IRQ>
    inline static void disable() {
        Control::template enable<IRQ>::type::apply(0);
    }

    template<const id_t& IRQ>
    inline static void enable() {
        Control::template enable<IRQ>::type::apply(1);
    }

    //--------------------------------------------------------------------------
    template<const id_t& IRQ,
             const typename trigger::id_t& Trigger>
    inline static void set_trigger() {
        Control::template set_trigger<IRQ>::type::apply(*Trigger);
    }

    //--------------------------------------------------------------------------
    inline static void enable_global()  { sei(); }
    inline static void disable_global() { cli(); }

    //--------------------------------------------------------------------------
    template<const id_t& IRQ>
    inline static void attach(isr_t isr) {
        mc::detail::isr_table::value[*IRQ] = isr;
    }
};

//------------------------------------------------------------------------------
#define MC_USE_DYNAMIC_IRQS()                                                  \
    mc::device::irq::isr_t                                                     \
        mc::detail::isr_table::value[mc::device::irq::count];

#define _MC_DI_CAT(X, ...) X ## __VA_ARGS__

#define MC_ENABLE_DYNAMIC_IRQ(NR)                                              \
    extern "C" void _MC_DI_CAT(__vector_,NR) () __attribute__ ((signal, used));\
    void _MC_DI_CAT(__vector_,NR) () { mc::detail::isr_table::value[NR](); }
