struct basic_masks {

//------------------------------------------------------------------------------
    template<const reg::id_t& R, class... Cs> using W = bit::write<R,Cs...>;
    template<count_t  B > using S = bit::set<B>;
    template<count_t  B > using C = bit::clr<B>;

//------------------------------------------------------------------------------
    template<const reg::id_t& R, count_t... Bs> using R = bit::read<R,Bs...>;
    template<const reg::id_t& R, count_t... Bs> using P = bit::put <R,Bs...>;

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg, count_t Bit> using set = W<Reg, S<Bit>>;
    template<const reg::id_t& Reg, count_t Bit> using clr = W<Reg, C<Bit>>;
    template<const reg::id_t& Reg, count_t Bit> using get = R<Reg, Bit>;

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg, mc::word_t Mask>
    struct write_word {
        inline static void apply(const mc::word_t n) {
            reg::write<Reg>((reg::read<Reg>() & ~Mask) | (n & Mask));
        }
    };

    template<const reg::id_t& Reg, mc::word_t Mask>
    struct read_word {
        inline static mc::word_t apply() {
            return reg::read<Reg>() & Mask;
        }
    };

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg>
    struct write_reg {
        inline static void apply(const mc::word_t n) {
            reg::write<Reg>(n);
        }
    };

    template<const reg::id_t& Reg>
    struct read_reg {
        inline static mc::word_t apply() {
            return reg::read<Reg>();
        }
    };

    template<const reg::id_t& Reg>
    struct rw_reg {
        using read = read_reg<Reg>;
        using write = write_reg<Reg>;
    };

//------------------------------------------------------------------------------
    template<const reg::id_t& RegHigh, const reg::id_t& RegLow>
    struct write_dreg {
        inline static void apply(const mc::dword_t n) {
            reg::write<RegHigh>(n >> mc::dword_half_offset);
            reg::write<RegLow >(n &  mc::dword_half_cut   );
        }
    };

    template<const reg::id_t& RegHigh, const reg::id_t& RegLow>
    struct read_dreg {
        inline static mc::dword_t apply() {
            return (reg::read<RegHigh>() << mc::dword_half_offset) |
                   (reg::read<RegLow>()                          ) ;
        }
    };

    template<const reg::id_t& RegHigh, const reg::id_t& RegLow>
    struct rw_dreg {
        using read = read_dreg<RegHigh, RegLow>;
        using write = write_dreg<RegHigh, RegLow>;
    };

//------------------------------------------------------------------------------
    struct undefined {};

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg, count_t B0>
    struct mode_1 {
        inline static void apply(const count_t n) {
            switch(n) {
                case 0: W<Reg, C<B0>>::apply(); break;
                case 1: W<Reg, S<B0>>::apply(); break;
            }
        }
    };

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg, count_t B1, count_t B0>
    struct mode_2 {
        inline static void apply(const count_t n) {
            switch(n) {
                case 0: W<Reg, C<B1>, C<B0>>::apply(); break;
                case 1: W<Reg, C<B1>, S<B0>>::apply(); break;
                case 2: W<Reg, S<B1>, C<B0>>::apply(); break;
                case 3: W<Reg, S<B1>, S<B0>>::apply(); break;
            }
        }
    };

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg, count_t B2, count_t B1, count_t B0>
    struct mode_3 {
        inline static void apply(const count_t n) {
            switch(n) {
                case 0: W<Reg, C<B2>, C<B1>, C<B0>>::apply(); break;
                case 1: W<Reg, C<B2>, C<B1>, S<B0>>::apply(); break;
                case 2: W<Reg, C<B2>, S<B1>, C<B0>>::apply(); break;
                case 3: W<Reg, C<B2>, S<B1>, S<B0>>::apply(); break;
                case 4: W<Reg, S<B2>, C<B1>, C<B0>>::apply(); break;
                case 5: W<Reg, S<B2>, C<B1>, S<B0>>::apply(); break;
                case 6: W<Reg, S<B2>, S<B1>, C<B0>>::apply(); break;
                case 7: W<Reg, S<B2>, S<B1>, S<B0>>::apply(); break;
            }
        }
    };

//------------------------------------------------------------------------------
    template<const reg::id_t& Reg,
             count_t B3, count_t B2, count_t B1, count_t B0>
    struct mode_4 {
        inline static void apply(const count_t n) {
            switch(n) {
                case 0:  W<Reg, C<B3>, C<B2>, C<B1>, C<B0>>::apply(); break;
                case 1:  W<Reg, C<B3>, C<B2>, C<B1>, S<B0>>::apply(); break;
                case 2:  W<Reg, C<B3>, C<B2>, S<B1>, C<B0>>::apply(); break;
                case 3:  W<Reg, C<B3>, C<B2>, S<B1>, S<B0>>::apply(); break;
                case 4:  W<Reg, C<B3>, S<B2>, C<B1>, C<B0>>::apply(); break;
                case 5:  W<Reg, C<B3>, S<B2>, C<B1>, S<B0>>::apply(); break;
                case 6:  W<Reg, C<B3>, S<B2>, S<B1>, C<B0>>::apply(); break;
                case 7:  W<Reg, C<B3>, S<B2>, S<B1>, S<B0>>::apply(); break;
                case 8:  W<Reg, S<B3>, C<B2>, C<B1>, C<B0>>::apply(); break;
                case 9:  W<Reg, S<B3>, C<B2>, C<B1>, S<B0>>::apply(); break;
                case 10: W<Reg, S<B3>, C<B2>, S<B1>, C<B0>>::apply(); break;
                case 11: W<Reg, S<B3>, C<B2>, S<B1>, S<B0>>::apply(); break;
                case 12: W<Reg, S<B3>, S<B2>, C<B1>, C<B0>>::apply(); break;
                case 13: W<Reg, S<B3>, S<B2>, C<B1>, S<B0>>::apply(); break;
                case 14: W<Reg, S<B3>, S<B2>, S<B1>, C<B0>>::apply(); break;
                case 15: W<Reg, S<B3>, S<B2>, S<B1>, S<B0>>::apply(); break;
            }
        }
    };

//------------------------------------------------------------------------------
    // this represents a wacky mode in which the bits are split between two
    // registers. (see atmega32 timer waveforms)
    template<const reg::id_t& RegHigh, count_t B2,
             const reg::id_t& RegLow, count_t B1, count_t B0>
    struct mode_2_1 { // TODO rename mode_1_2
        inline static void apply(const count_t n) {
            switch(n) {
                case 0: W<RegHigh, C<B2>       >::apply();
                        W<RegLow , C<B1>, C<B0>>::apply(); break;

                case 1: W<RegHigh, C<B2>       >::apply();
                        W<RegLow , C<B1>, S<B0>>::apply(); break;

                case 2: W<RegHigh, C<B2>       >::apply();
                        W<RegLow , S<B1>, C<B0>>::apply(); break;

                case 3: W<RegHigh, C<B2>       >::apply();
                        W<RegLow , S<B1>, S<B0>>::apply(); break;

                case 4: W<RegHigh, S<B2>       >::apply();
                        W<RegLow , C<B1>, C<B0>>::apply(); break;

                case 5: W<RegHigh, S<B2>       >::apply();
                        W<RegLow , C<B1>, S<B0>>::apply(); break;

                case 6: W<RegHigh, S<B2>       >::apply();
                        W<RegLow , S<B1>, C<B0>>::apply(); break;

                case 7: W<RegHigh, S<B2>       >::apply();
                        W<RegLow , S<B1>, S<B0>>::apply(); break;
            }
        }
    };

//------------------------------------------------------------------------------
    // another wacky mode in which the bits are split between two
    // registers. (see atmega32 timer waveforms)
    template<const reg::id_t& RegHigh, count_t B3, count_t B2,
             const reg::id_t& RegLow,  count_t B1, count_t B0>
    struct mode_2_2 {
        inline static void apply(const count_t n) {
            switch(n) {
                case 0:  W<RegHigh, C<B3>, C<B2>>::apply();
                         W<RegLow , C<B1>, C<B0>>::apply(); break;

                case 1:  W<RegHigh, C<B3>, C<B2>>::apply();
                         W<RegLow , C<B1>, S<B0>>::apply(); break;

                case 2:  W<RegHigh, C<B3>, C<B2>>::apply();
                         W<RegLow , S<B1>, C<B0>>::apply(); break;

                case 3:  W<RegHigh, C<B3>, C<B2>>::apply();
                         W<RegLow , S<B1>, S<B0>>::apply(); break;

                case 4:  W<RegHigh, C<B3>, S<B2>>::apply();
                         W<RegLow , C<B1>, C<B0>>::apply(); break;

                case 5:  W<RegHigh, C<B3>, S<B2>>::apply();
                         W<RegLow , C<B1>, S<B0>>::apply(); break;

                case 6:  W<RegHigh, C<B3>, S<B2>>::apply();
                         W<RegLow , S<B1>, C<B0>>::apply(); break;

                case 7:  W<RegHigh, C<B3>, S<B2>>::apply();
                         W<RegLow , S<B1>, S<B0>>::apply(); break;

                case 8:  W<RegHigh, S<B3>, C<B2>>::apply();
                         W<RegLow , C<B1>, C<B0>>::apply(); break;

                case 9:  W<RegHigh, S<B3>, C<B2>>::apply();
                         W<RegLow , C<B1>, S<B0>>::apply(); break;

                case 10: W<RegHigh, S<B3>, C<B2>>::apply();
                         W<RegLow , S<B1>, C<B0>>::apply(); break;

                case 11: W<RegHigh, S<B3>, C<B2>>::apply();
                         W<RegLow , S<B1>, S<B0>>::apply(); break;

                case 12: W<RegHigh, S<B3>, S<B2>>::apply();
                         W<RegLow , C<B1>, C<B0>>::apply(); break;

                case 13: W<RegHigh, S<B3>, S<B2>>::apply();
                         W<RegLow , C<B1>, S<B0>>::apply(); break;

                case 14: W<RegHigh, S<B3>, S<B2>>::apply();
                         W<RegLow , S<B1>, C<B0>>::apply(); break;

                case 15: W<RegHigh, S<B3>, S<B2>>::apply();
                         W<RegLow , S<B1>, S<B0>>::apply(); break;
            }
        }
    };


//------------------------------------------------------------------------------
};
