//------------------------------------------------------------------------------
// BASIC PINSET
//------------------------------------------------------------------------------
// for some reason (that eludes me) this has to be defined outside the
// pins struct.
struct basic_pin_id : public value_enum<count_t> {
        constexpr basic_pin_id(count_t id,
                               count_t bit,
                               const basic_register_set::id_t& ddr,
                               const basic_register_set::id_t& port,
                               const basic_register_set::id_t& pin)
        : value_enum(id),
          bit(bit),
          reg_ddr(ddr), reg_port(port), reg_pin(pin) {}

    count_t bit;
    const basic_register_set::id_t& reg_ddr;
    const basic_register_set::id_t& reg_port;
    const basic_register_set::id_t& reg_pin;
};

//------------------------------------------------------------------------------
template<class Reg>
struct basic_pin_set {

    using id_t = basic_pin_id;

    //--------------------------------------------------------------------------
    enum class mode  { input, output };
    enum class level { low, high     };

    //--------------------------------------------------------------------------
    constexpr static level inverse_level(level l) {
        return l == level::high ? level::low : level::high;
    }

    //--------------------------------------------------------------------------
    template<const basic_pin_id& Pin, pin_register_kind Kind>
    struct lookup {
        constexpr static const basic_register_set::id_t* value = 
              (Kind == pin_register_kind::direction ? &Pin.reg_ddr
            : (Kind == pin_register_kind::pin       ? &Pin.reg_pin
            :                                         &Pin.reg_port));
    };

    //--------------------------------------------------------------------------
    template<pin_register_kind Kind, const id_t& Pin>
    inline static void set_reg() {
        Reg::template set<*lookup<Pin,Kind>::value, Pin.bit>();
    }

    template<pin_register_kind Kind, const id_t& Pin>
    inline static void clr_reg() {
        Reg::template clr<*lookup<Pin, Kind>::value, Pin.bit>();
    }
    
    template<pin_register_kind Kind, const id_t& Pin>
    inline static bool get_reg() {
        return Reg::template get<*lookup<Pin,Kind>::value, Pin.bit>();
    }

    //--------------------------------------------------------------------------
    template<const id_t& Pin>
    inline static void set(level l) {
        if (l == level::low) {
            clr_reg<pin_register_kind::port, Pin>();
        } else {
            set_reg<pin_register_kind::port, Pin>();
        }
    }

    template<const id_t& Pin>
    inline static void set(bool l) {
        return l ? set<Pin>(level::high)
                 : set<Pin>(level::low);
    }

    //--------------------------------------------------------------------------
    template<const id_t& Pin>
    inline static level get() {
        return get_reg<pin_register_kind::pin, Pin>()
                ? level::high 
                : level::low;
    }

    //--------------------------------------------------------------------------
    template<const id_t& Pin>
    inline static void configure(mode m) {
        if (m == mode::input) {
            clr_reg<pin_register_kind::direction, Pin>();
        } else {
            set_reg<pin_register_kind::direction, Pin>();
        }
    }
};
