//------------------------------------------------------------------------------
// BASIC SPI
//------------------------------------------------------------------------------

template<
    class Control,
    class Status,
    class Data,
    class Pin
> struct basic_spi {

    //--------------------------------------------------------------------------
    using mode      = typename Control::mode;
    using order     = typename Control::data_order;
    using polarity  = typename Control::clock_polarity;
    using phase     = typename Control::clock_phase;
    using prescaler = typename Control::clock_prescaler;
    using pin       = Pin;

    //--------------------------------------------------------------------------
    inline static void enable()  { Control::enable::apply(1); }
    inline static void disable() { Control::enable::apply(0); }

    //--------------------------------------------------------------------------
    template<const typename order::id_t& DO>
    inline static void set_order() {
        order::set::apply(*DO);
    }

    //--------------------------------------------------------------------------
    template<const typename mode::id_t& M>
    inline static void set_mode() {
        mode::set::apply(*M);
    }

    //--------------------------------------------------------------------------
    template<const typename polarity::id_t& P>
    inline static void set_polarity() {
        polarity::set::apply(*P);
    }

    template<const typename phase::id_t& P>
    inline static void set_phase() {
        phase::set::apply(*P);
    }
    
    template<const typename prescaler::id_t& P>
    inline static void set_prescaler() {
        prescaler::set::apply(*P);
    }

    //--------------------------------------------------------------------------
    inline static bool transfer_complete() {
        return Status::transfer_complete::apply();
    }

    inline static bool write_collision() {
        return Status::write_collision::apply();
    }

    //--------------------------------------------------------------------------
    inline static word_t read() {
        return Data::read::apply();
    }

    inline static void write(word_t d) {
        Data::write::apply(d);
    }
};
