//------------------------------------------------------------------------------
// BASIC TIMER
//------------------------------------------------------------------------------
template<count_t Resolution,
         class Clock,    class ClockMask,
         class Waveform, class WaveformMask,
         class PWM,      template<const basic_pin_id&> class  PWMPinModeLookup,
         class RegisterAccess>
struct basic_timer {

    constexpr static count_t resolution = Resolution;

    using pwm           = PWM;
    using clock         = Clock;
    using clock_mask    = ClockMask;
    using waveform      = Waveform;
    using waveform_mask = WaveformMask;
    using reg_value_t   = typename RegisterAccess::value_t;

    //--------------------------------------------------------------------------
    template<const typename clock::id_t& Clk>
    inline static void set_clock() {
        clock_mask::apply(*Clk);
    }

    //--------------------------------------------------------------------------
    template<const typename waveform::id_t& Wave>
    inline static void set_waveform() {
        waveform_mask::apply(*Wave);
    }

    //--------------------------------------------------------------------------
    template<const basic_pin_id& Pin, const typename pwm::id_t& Mode>
    inline static void set_pwm_mode() {
        PWMPinModeLookup<Pin>::mask::apply(*Mode);
    }

    //--------------------------------------------------------------------------
    inline static void set_cnt(const reg_value_t v) {
        RegisterAccess::cnt::write::apply(v);
    }

    inline static reg_value_t get_cnt() {
        return RegisterAccess::cnt::read::apply();
    }

    //--------------------------------------------------------------------------
    inline static void set_ocra(const reg_value_t v) {
        RegisterAccess::ocra::write::apply(v);
    }

    inline static reg_value_t get_ocra() {
        return RegisterAccess::ocra::read::apply();
    }

    inline static void set_ocrb(const reg_value_t v) {
        RegisterAccess::ocrb::write::apply(v);
    }

    inline static reg_value_t get_ocrb() {
        return RegisterAccess::ocrb::read::apply();
    }

    //--------------------------------------------------------------------------
    inline static void set_icr(const reg_value_t v) {
        RegisterAccess::icr::write::apply(v);
    }

    inline static reg_value_t get_icr() {
        return RegisterAccess::icr::read::apply();
    }
};
