//------------------------------------------------------------------------------
// BASIC TWI
//------------------------------------------------------------------------------
template<
    class Frequency,
    class Control,
    class ReadStatus,
    class DataAccess,
    class SlaveAddress,
    class Mode
> struct basic_twi {

    //--------------------------------------------------------------------------
    using mode   = Mode;
    using freq_t = typename Frequency::freq_t;

    //--------------------------------------------------------------------------
    template<freq_t                      Freq,
             typename Frequency::value_t Rate = Frequency::calc_rate(Freq)>
    inline static void set_frequency() {
        static_assert(Rate >= Frequency::min_rate,
                      "twi frequency too high.");

        Frequency::set_rate::apply(Rate);
    }

    //--------------------------------------------------------------------------
    inline static void enable() {
        Control::enable::apply(1);
    }

    inline static void disable() {
        Control::enable::apply(0);
    }

    //--------------------------------------------------------------------------
    inline static void set_slave_address(mc::word_t addr,
                                  bool with_general_call = false) {
        SlaveAddress::apply(addr | (with_general_call ? 1 : 0));
    }

    inline static void set_slave_ready() {
        Control::slave_ready::apply();
    }

    //--------------------------------------------------------------------------
    inline static void start() {
        Control::start::apply();
    }

    inline static void stop() {
        Control::stop::apply();
    }

    //--------------------------------------------------------------------------
    inline static mc::word_t status() {
        return ReadStatus::apply();
    }

    //--------------------------------------------------------------------------
    inline static mc::word_t read() {
        return DataAccess::read::apply();
    }

    //--------------------------------------------------------------------------
    inline static void write(mc::word_t w) {
        DataAccess::write::apply(w);
    }

    //--------------------------------------------------------------------------
    template<const typename Mode::id_t& M>
    inline static void write_address(mc::word_t addr) {
        write((addr << 1) | *M);
    }

    //--------------------------------------------------------------------------
    inline static void step() {
        Control::step::apply();
    }

    inline static bool transmit_complete() {
        return Control::transmit_complete::apply();
    };

    inline static bool stop_complete() {
        return !Control::stop_pending::apply();
    };

    //--------------------------------------------------------------------------
    inline static void ack() {
        Control::ack::apply();
    }

    inline static void nack() {
        Control::nack::apply();
    }
};
