//------------------------------------------------------------------------------
// BASIC UART
//------------------------------------------------------------------------------
template<
    class BaudControl,
    class StatusInformation,
    class Control,
    class Data
> struct basic_uart {

    //--------------------------------------------------------------------------
    using baud    = BaudControl;
    using status  = StatusInformation;
    using control = Control;
    using data    = Data;

    //--------------------------------------------------------------------------
    inline static typename Data::value_t read() {
        return Data::read::apply();
    }

    inline static void write(typename Data::value_t v) {
        return Data::write::apply(v);
    }

    //--------------------------------------------------------------------------
    inline void putchar(char c) {
        while(! data_register_empty()) {}
        write(c);
    }

    inline char getchar() {
        while(! recv_complete()) {}
        return read();
    }

    //--------------------------------------------------------------------------
    template<typename BaudControl::baud_t  Baud,
             typename BaudControl::value_t BRR   = BaudControl::calc_rate(Baud),
             typename BaudControl::value_t Error
                        = BaudControl::calc_error(Baud, BRR)>
    inline static void set_baud_rate() {
        static_assert(Error >= BaudControl::min_error &&
                      Error <= BaudControl::max_error,
                      "baud rate error too high.");

        BaudControl::set_rate::apply(BRR);
    }

    //--------------------------------------------------------------------------
    // status flags
    inline static bool recv_complete() {
        return StatusInformation::recv_complete::apply();
    }

    inline static bool transmit_complete() {
        return StatusInformation::transmit_complete::apply();
    }

    inline static bool data_register_empty() {
        return StatusInformation::data_register_empty::apply();
    }

    inline static bool framing_error() {
        return StatusInformation::framing_error::apply();
    }

    inline static bool data_overrun() {
        return StatusInformation::data_overrun::apply();
    }

    inline static bool parity_error() {
        return StatusInformation::parity_error::apply();
    }

    //--------------------------------------------------------------------------
    // control
    inline static void set_recv(bool b) {
        Control::recv_enable::apply(b ? 1 : 0);
    }

    inline static void set_transmit(bool b) {
        Control::transmit_enable::apply(b ? 1 : 0);
    }

    template<const typename Control::mode::id_t& M>
    inline static void set_mode() {
        Control::select_mode::apply(*M);
    }

    template<const typename Control::stop_bits::id_t& M>
    inline static void set_stop_bits() {
        Control::select_stop_bits::apply(*M);
    }

    template<const typename Control::parity::id_t& M>
    inline static void set_parity() {
        Control::select_parity::apply(*M);
    }

    template<const typename Control::character_size::id_t& M>
    inline static void set_character_size() {
        Control::select_character_size::apply(*M);
    }
};
