#pragma once

#define MC_CPU ATMEGA328P

#include <stdint.h>
namespace mc { namespace detail {
    struct cpu_conf {
        using word_t                                = uint8_t;
        using dword_t                               = uint16_t;
        constexpr static uint8_t  dword_half_offset = 8;
        constexpr static uint16_t dword_half_cut    = 0xFF;
    };
}}

#include "../common/value_enum.hpp"
#include "../common/header.inc.hpp"

//------------------------------------------------------------------------------
namespace mc { namespace detail {

//------------------------------------------------------------------------------
// REGISTERS
//------------------------------------------------------------------------------
struct reg : public basic_register_set {
    #define io_ref(ADDRESS) (ADDRESS + __SFR_OFFSET)
    #define    ref(ADDRESS) (ADDRESS)

//------------------------------------------------------------------------------
// PINS
    constexpr static id_t pinb   = io_ref(0x03);
    constexpr static id_t ddrb   = io_ref(0x04);
    constexpr static id_t portb  = io_ref(0x05);

    constexpr static id_t pinc   = io_ref(0x06);
    constexpr static id_t ddrc   = io_ref(0x07);
    constexpr static id_t portc  = io_ref(0x08);

    constexpr static id_t pind   = io_ref(0x09);
    constexpr static id_t ddrd   = io_ref(0x0A);
    constexpr static id_t portd  = io_ref(0x0B);

//------------------------------------------------------------------------------
// INTERRUPTS
    constexpr static id_t eicra  =    ref(0x69);
    constexpr static id_t eimsk  = io_ref(0x1D);

//------------------------------------------------------------------------------
// ADC
    constexpr static id_t adcsra =    ref(0x7A);
    constexpr static id_t admux  =    ref(0x7C);
    constexpr static id_t adcl   =    ref(0x78);
    constexpr static id_t adch   =    ref(0x79);

//------------------------------------------------------------------------------
// TIMERS
    constexpr static id_t tccr0a = io_ref(0x24);
    constexpr static id_t tccr0b = io_ref(0x25);
    constexpr static id_t tccr1a =    ref(0x80);
    constexpr static id_t tccr1b =    ref(0x81);
    constexpr static id_t tccr1c =    ref(0x82);
    constexpr static id_t tccr2a =    ref(0xB0);
    constexpr static id_t tccr2b =    ref(0xB1);

    constexpr static id_t timsk0 =    ref(0x6E);
    constexpr static id_t timsk1 =    ref(0x6F);
    constexpr static id_t timsk2 =    ref(0x70);

    constexpr static id_t tcnt0  = io_ref(0x26);
    constexpr static id_t tcnt1l =    ref(0x84);
    constexpr static id_t tcnt1h =    ref(0x85);
    constexpr static id_t tcnt2  =    ref(0xB2);

    constexpr static id_t ocr0a  = io_ref(0x27);
    constexpr static id_t ocr0b  = io_ref(0x28);
    constexpr static id_t ocr1al =    ref(0x88);
    constexpr static id_t ocr1ah =    ref(0x89);
    constexpr static id_t ocr1bl =    ref(0x8A);
    constexpr static id_t ocr1bh =    ref(0x8B);
    constexpr static id_t ocr2a  =    ref(0xB3);
    constexpr static id_t ocr2b  =    ref(0xB4);

    constexpr static id_t icr1l  =    ref(0x86);
    constexpr static id_t icr1h  =    ref(0x87);

//------------------------------------------------------------------------------
// UART
    constexpr static id_t ucsr0a =    ref(0xC0);
    constexpr static id_t ucsr0b =    ref(0xC1);
    constexpr static id_t ucsr0c =    ref(0xC2);

    constexpr static id_t ubrr0l =    ref(0xC4);
    constexpr static id_t ubrr0h =    ref(0xC5);

    constexpr static id_t udr0   =    ref(0xC6);

//------------------------------------------------------------------------------
// TWI
    constexpr static id_t twbr   =    ref(0xB8);
    constexpr static id_t twsr   =    ref(0xB9);
    constexpr static id_t twar   =    ref(0xBA);
    constexpr static id_t twdr   =    ref(0xBB);
    constexpr static id_t twcr   =    ref(0xBC);

//------------------------------------------------------------------------------
// SPI
    constexpr static id_t spcr   = io_ref(0x2C);
    constexpr static id_t spsr   = io_ref(0x2D);
    constexpr static id_t spdr   = io_ref(0x2E);
};


//------------------------------------------------------------------------------
// PINS
//------------------------------------------------------------------------------
struct pin : public basic_pin_set<reg> {

    constexpr static id_t pb0 = {0x00, 0, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb1 = {0x01, 1, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb2 = {0x02, 2, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb3 = {0x03, 3, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb4 = {0x04, 4, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb5 = {0x05, 5, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb6 = {0x06, 6, reg::ddrb, reg::portb, reg::pinb};
    constexpr static id_t pb7 = {0x07, 7, reg::ddrb, reg::portb, reg::pinb};

    constexpr static id_t pc0 = {0x08, 0, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc1 = {0x09, 1, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc2 = {0x0A, 2, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc3 = {0x0B, 3, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc4 = {0x0C, 4, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc5 = {0x0D, 5, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc6 = {0x0E, 6, reg::ddrc, reg::portc, reg::pinc};
    constexpr static id_t pc7 = {0x0F, 7, reg::ddrc, reg::portc, reg::pinc};

    constexpr static id_t pd0 = {0x10, 0, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd1 = {0x11, 1, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd2 = {0x12, 2, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd3 = {0x13, 3, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd4 = {0x14, 4, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd5 = {0x15, 5, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd6 = {0x16, 6, reg::ddrd, reg::portd, reg::pind};
    constexpr static id_t pd7 = {0x17, 7, reg::ddrd, reg::portd, reg::pind};

    struct pwm {
        constexpr static const id_t oc0a = pd6;
        constexpr static const id_t oc0b = pd5;
        constexpr static const id_t oc1a = pb1;
        constexpr static const id_t oc1b = pb2;
        constexpr static const id_t oc2a = pb3;
        constexpr static const id_t oc2b = pd3;
    };

    struct spi {
        constexpr static const id_t mosi = pb3;
        constexpr static const id_t miso = pb4;
        constexpr static const id_t sck  = pb5;
    };
};

//------------------------------------------------------------------------------
// BIT MASKS
//------------------------------------------------------------------------------

#include "../common/bit_mask.inc.hpp"
#include "../common/masks.inc.hpp"

struct masks : public basic_masks {
    //--------------------------------------------------------------------------
    template<const reg::id_t& CTRL, count_t S02, count_t S01, count_t S00>
    using timer_set_clock = mode_3<CTRL, S02, S01, S00>;
};

//------------------------------------------------------------------------------
// INTERRUPTS
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
struct irq_control {
    template<const basic_irq_id& IRQ> struct set_trigger;
    template<const basic_irq_id& Irq> struct enable;
};

//------------------------------------------------------------------------------
struct irq : public basic_irq <
    irq_control,
    26
> {
    constexpr static id_t int0 = INT0_vect_num;
    constexpr static id_t int1 = INT1_vect_num;
    constexpr static id_t adc  = ADC_vect_num;

    constexpr static id_t timer0_compa = TIMER0_COMPA_vect_num;
    constexpr static id_t timer0_compb = TIMER0_COMPB_vect_num;
    constexpr static id_t timer0_ovf   = TIMER0_OVF_vect_num;

    constexpr static id_t timer1_capt  = TIMER1_CAPT_vect_num;
    constexpr static id_t timer1_compa = TIMER1_COMPA_vect_num;
    constexpr static id_t timer1_compb = TIMER1_COMPB_vect_num;
    constexpr static id_t timer1_ovf   = TIMER1_OVF_vect_num;

    constexpr static id_t timer2_compa = TIMER2_COMPA_vect_num;
    constexpr static id_t timer2_compb = TIMER2_COMPB_vect_num;
    constexpr static id_t timer2_ovf   = TIMER2_OVF_vect_num;

    constexpr static id_t usart_rx     = USART_RX_vect_num;
    constexpr static id_t usart_udre   = USART_UDRE_vect_num;
    constexpr static id_t usart_tx     = USART_TX_vect_num;

    constexpr static id_t spi_stc      = SPI_STC_vect_num;
};

//------------------------------------------------------------------------------
template<> struct irq_control::set_trigger<irq::int0> {
    using type = masks::mode_2<reg::eicra, ISC01, ISC00>;
};

template<> struct irq_control::set_trigger<irq::int1> {
    using type = masks::mode_2<reg::eicra, ISC11, ISC10>;
};

//------------------------------------------------------------------------------
#define IRQ_SWITCH(IRQ, REGISTER, BIT)                                         \
    template<> struct irq_control::enable< IRQ >                               \
        { using type = masks::mode_1< REGISTER, BIT >; };
IRQ_SWITCH(irq::int0        , reg::eimsk , INT0  )
IRQ_SWITCH(irq::int1        , reg::eimsk , INT1  )
IRQ_SWITCH(irq::timer0_ovf  , reg::timsk0, TOIE0 )
IRQ_SWITCH(irq::timer0_compa, reg::timsk0, OCIE0A)
IRQ_SWITCH(irq::timer0_compb, reg::timsk0, OCIE0B)
IRQ_SWITCH(irq::timer1_ovf  , reg::timsk1, TOIE1 )
IRQ_SWITCH(irq::timer1_compa, reg::timsk1, OCIE1A)
IRQ_SWITCH(irq::timer1_compb, reg::timsk1, OCIE1B)
IRQ_SWITCH(irq::timer1_capt , reg::timsk1, ICIE1 )
IRQ_SWITCH(irq::timer2_ovf  , reg::timsk2, TOIE2 )
IRQ_SWITCH(irq::timer2_compa, reg::timsk2, OCIE2A)
IRQ_SWITCH(irq::timer2_compb, reg::timsk2, OCIE2B)
IRQ_SWITCH(irq::usart_rx    , reg::ucsr0b, RXCIE0)
IRQ_SWITCH(irq::usart_udre  , reg::ucsr0b, UDRIE0)
IRQ_SWITCH(irq::usart_tx    , reg::ucsr0b, TXCIE0)
IRQ_SWITCH(irq::spi_stc     , reg::spcr  , SPIE  )
IRQ_SWITCH(irq::adc         , reg::adcsra, ADIE  )
#undef IRQ_SWITCH

//------------------------------------------------------------------------------
// TIMERS
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
struct clock_source_01 {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t stop    = 0;
    constexpr static id_t ps1     = 1;
    constexpr static id_t ps8     = 2;
    constexpr static id_t ps64    = 3;
    constexpr static id_t ps256   = 4;
    constexpr static id_t ps1024  = 5;
    constexpr static id_t falling = 6;
    constexpr static id_t rising  = 7;
};

struct clock_source_2 {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t stop    = 0;
    constexpr static id_t ps1     = 1;
    constexpr static id_t ps8     = 2;
    constexpr static id_t ps32    = 3;
    constexpr static id_t ps64    = 4;
    constexpr static id_t ps128   = 5;
    constexpr static id_t ps256   = 6;
    constexpr static id_t ps1024  = 7;
};

//------------------------------------------------------------------------------
struct pwm_mode {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t disabled = 0;
    constexpr static id_t toggle   = 1;
    constexpr static id_t normal   = 2;
    constexpr static id_t inverted = 3;
};

//------------------------------------------------------------------------------
template<const pin::id_t&> struct pwm_pin_mode_lookup {};

template<> struct pwm_pin_mode_lookup<pin::pwm::oc0a> {
    using mask = masks::mode_2<reg::tccr0a, COM0A1, COM0A0>;
};
template<> struct pwm_pin_mode_lookup<pin::pwm::oc0b> {
    using mask = masks::mode_2<reg::tccr0a, COM0B1, COM0B0>;
};
template<> struct pwm_pin_mode_lookup<pin::pwm::oc1a> {
    using mask = masks::mode_2<reg::tccr1a, COM1A1, COM1A0>;
};
template<> struct pwm_pin_mode_lookup<pin::pwm::oc1b> {
    using mask = masks::mode_2<reg::tccr1a, COM1B1, COM1B0>;
};
template<> struct pwm_pin_mode_lookup<pin::pwm::oc2a> {
    using mask = masks::mode_2<reg::tccr2a, COM2A1, COM2A0>;
};
template<> struct pwm_pin_mode_lookup<pin::pwm::oc2b> {
    using mask = masks::mode_2<reg::tccr2a, COM2B1, COM2B0>;
};

//------------------------------------------------------------------------------
struct waveform_timer0 {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t normal          = 0;
    constexpr static id_t pwm_pc          = 1;
    constexpr static id_t ctc             = 2;
    constexpr static id_t fast_pwm        = 3;
    constexpr static id_t pwm_pc_toggle   = 5;
    constexpr static id_t fast_pwm_toggle = 7;
};

struct register_access_timer0 {
    using value_t = word_t;
    using cnt     = masks::rw_reg<reg::tcnt0>;
    using ocra    = masks::rw_reg<reg::ocr0a>;
    using ocrb    = masks::rw_reg<reg::ocr0b>;
    using icr     = masks::undefined;
};

struct timer0 : basic_timer<
    8,
    clock_source_01, masks::timer_set_clock<reg::tccr0b, CS02, CS01, CS00>,
    waveform_timer0, masks::mode_2_1<reg::tccr0b, WGM02,
                                     reg::tccr0a, WGM01, WGM00>,
    pwm_mode, pwm_pin_mode_lookup,
    register_access_timer0
> {
    struct defaults {
        struct pwm {
            constexpr static const auto* Clk  = &clock::ps64;
            constexpr static const auto* Wave = &waveform::fast_pwm;
            constexpr static const auto* Mode = &pwm_mode::normal;
        };
    };
};

//------------------------------------------------------------------------------
struct waveform_timer1 {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t normal       = 0;
    constexpr static id_t pwm_pc8      = 1;
    constexpr static id_t pwm_pc9      = 2;
    constexpr static id_t pwm_pc10     = 3;
    constexpr static id_t ctc_ocr      = 4;
    constexpr static id_t fast_pwm8    = 5;
    constexpr static id_t fast_pwm9    = 6;
    constexpr static id_t fast_pwm10   = 7;
    constexpr static id_t pwm_pfc_icr  = 8;
    constexpr static id_t pwm_pfc_ocr  = 9;
    constexpr static id_t pwm_pc_icr   = 10;
    constexpr static id_t pwm_pc_ocr   = 11;
    constexpr static id_t ctc_icr      = 12;
    constexpr static id_t fast_pwm_icr = 14;
    constexpr static id_t fast_pwm_ocr = 15;
};

struct register_access_timer1 {
    using value_t = dword_t;
    using cnt     = masks::rw_dreg<reg::tcnt1h, reg::tcnt1l>;
    using ocra    = masks::rw_dreg<reg::ocr1ah, reg::ocr1al>;
    using ocrb    = masks::rw_dreg<reg::ocr1bh, reg::ocr1bl>;
    using icr     = masks::rw_dreg<reg::icr1h , reg::icr1l >;
};

struct timer1 : basic_timer<
    16,
    clock_source_01, masks::timer_set_clock<reg::tccr1b, CS12, CS11, CS10>,
    waveform_timer0, masks::mode_2_2<reg::tccr1b, WGM13, WGM12,
                                     reg::tccr1a, WGM11, WGM10>,
    pwm_mode, pwm_pin_mode_lookup,
    register_access_timer1
> {
    struct defaults {
        struct pwm {
            constexpr static const auto* Clk  = &clock::ps64;
            constexpr static const auto* Wave = &waveform::fast_pwm;
            constexpr static const auto* Mode = &pwm_mode::normal;
        };
    };
};

//------------------------------------------------------------------------------
struct waveform_timer2 {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t normal          = 0;
    constexpr static id_t pwm_pc          = 1;
    constexpr static id_t ctc             = 2;
    constexpr static id_t fast_pwm        = 3;
    constexpr static id_t pwm_pc_toggle   = 5;
    constexpr static id_t fast_pwm_toggle = 7;
};

struct register_access_timer2 {
    using value_t = word_t;
    using cnt     = masks::rw_reg<reg::tcnt2>;
    using ocra    = masks::rw_reg<reg::ocr2a>;
    using ocrb    = masks::rw_reg<reg::ocr2b>;
    using icr     = masks::undefined;
};

struct timer2 : basic_timer<
    8,
    clock_source_2,  masks::timer_set_clock<reg::tccr2b, CS22, CS21, CS20>,
    waveform_timer2, masks::mode_2_1<reg::tccr2b, WGM22,
                                     reg::tccr2a, WGM21, WGM20>,
    pwm_mode, pwm_pin_mode_lookup,
    register_access_timer2
> {
    struct defaults {
        struct pwm {
            constexpr static const auto* Clk  = &clock::ps32;
            constexpr static const auto* Wave = &waveform::fast_pwm;
            constexpr static const auto* Mode = &pwm_mode::normal;
        };
    };
};

//------------------------------------------------------------------------------
// ADC
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
struct adc_control {
    using Ms = masks;
    using read_start_conversion = bit::read<reg::adcsra, ADSC>;
    using set_resolution        = Ms::mode_1<reg::admux,  ADLAR              >;
    using set_reference         = Ms::mode_2<reg::admux , REFS1, REFS0       >;
    using set_prescaler         = Ms::mode_3<reg::adcsra, ADPS2, ADPS1, ADPS0>;
    using enable                = Ms::mode_1<reg::adcsra, ADEN>;
    using start                 = Ms::W<reg::adcsra, Ms::S<ADSC>>;
    using select_channel        = Ms::write_word<reg::admux, 0x1F>;

};

//------------------------------------------------------------------------------
struct adc_reference {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t external = 0;
    constexpr static id_t avcc     = 1;
    constexpr static id_t internal = 3;
};

//------------------------------------------------------------------------------
struct adc_prescaler {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t ps2   = 1;
    constexpr static id_t ps4   = 2;
    constexpr static id_t ps8   = 3;
    constexpr static id_t ps16  = 4;
    constexpr static id_t ps32  = 5;
    constexpr static id_t ps64  = 6;
    constexpr static id_t ps128 = 7;
};

//------------------------------------------------------------------------------
template<const pin::id_t& Pin>
struct adc_channel_lookup {
    constexpr static count_t value = *Pin - *pin::pc0;
};

//------------------------------------------------------------------------------
struct adc_resolution {
    struct r8 {
        using type = mc::word_t;
        constexpr static count_t value = 1;
        inline static type read() { return ADCH; }
    };

    struct r10 {
        using type = uint16_t;
        constexpr static count_t value = 0;
        inline static type read() { return ADCW; }
    };
};

//------------------------------------------------------------------------------
struct adc : public basic_adc <
    adc_control,
    adc_reference,
    adc_prescaler,
    adc_channel_lookup,
    adc_resolution
> {
    using r8  = adc_resolution::r8;
    using r10 = adc_resolution::r10;

    struct defaults {
        constexpr static const auto* reference  = &adc_reference::avcc;
        constexpr static const auto* prescaler  = &adc_prescaler::ps64;
        constexpr static const auto* irq        = &mc::detail::irq::adc;
        using                        resolution = adc_resolution::r10;
    };
};

//------------------------------------------------------------------------------
// UART
//------------------------------------------------------------------------------

//------------------------------------------------------------------------------
struct uart0_baud {
    using value_t = uint16_t;
    using baud_t  = uint32_t;

    constexpr static value_t calc_rate(const baud_t baud) {
        return (cpu_frequency + (baud * 8)) / (baud * 16) - 1;
    }

    constexpr static value_t calc_error(const baud_t baud, const value_t brr) {
        return ((cpu_frequency / (16 * (brr+1))) * 1000) / baud;
    }

    constexpr static value_t max_error = 1020;
    constexpr static value_t min_error = 980;

    using set_rate = masks::write_dreg<reg::ubrr0h, reg::ubrr0l>;
};

//------------------------------------------------------------------------------
struct uart0_status_information {
    using recv_complete       = masks::get<reg::ucsr0a, RXC0 >;
    using transmit_complete   = masks::get<reg::ucsr0a, TXC0 >;
    using data_register_empty = masks::get<reg::ucsr0a, UDRE0>;
    using framing_error       = masks::get<reg::ucsr0a, FE0  >;
    using data_overrun        = masks::get<reg::ucsr0a, DOR0 >;
    using parity_error        = masks::get<reg::ucsr0a, UPE0 >;
};

//------------------------------------------------------------------------------
struct uart0_control {
    using recv_enable      = masks::mode_1<reg::ucsr0b, RXEN0           >;
    using transmit_enable  = masks::mode_1<reg::ucsr0b, TXEN0           >;
    using select_mode      = masks::mode_2<reg::ucsr0c, UMSEL01, UMSEL00>;
    using select_parity    = masks::mode_2<reg::ucsr0c, UPM01  , UPM00  >;
    using select_stop_bits = masks::mode_1<reg::ucsr0c, USBS0           >;

    using select_character_size  = masks::mode_2_1<reg::ucsr0b, UCSZ02,
                                                   reg::ucsr0c, UCSZ01, UCSZ00>;

    struct mode {
        DEFINE_VALUE_ENUM(id_t, count_t)
        constexpr static id_t asynchronous = 0;
        constexpr static id_t synchronous  = 1;
        constexpr static id_t master_spi   = 3;
    };

    struct stop_bits {
        DEFINE_VALUE_ENUM(id_t, count_t)
        constexpr static id_t one = 0;
        constexpr static id_t two = 1;
    };

    struct parity {
        DEFINE_VALUE_ENUM(id_t, count_t)
        constexpr static id_t disabled = 0;
        constexpr static id_t even     = 0;
        constexpr static id_t odd      = 0;
    };

    struct character_size {
        DEFINE_VALUE_ENUM(id_t, count_t)
        constexpr static id_t cs5 = 0;
        constexpr static id_t cs6 = 1;
        constexpr static id_t cs7 = 2;
        constexpr static id_t cs8 = 3;
        // TODO: 9bit...?
    };
};

//------------------------------------------------------------------------------
struct uart0_data {
    using value_t = mc::word_t;
    using read    = masks::read_reg <reg::udr0>;
    using write   = masks::write_reg<reg::udr0>;
};

//------------------------------------------------------------------------------
struct uart0 : public basic_uart<
    uart0_baud,
    uart0_status_information,
    uart0_control,
    uart0_data
> {
    using mode           = uart0_control::mode;
    using stop_bits      = uart0_control::stop_bits;
    using parity         = uart0_control::parity;
    using character_size = uart0_control::character_size;

    struct defaults {
        constexpr static const auto  baud_rate = 9600;
        constexpr static const auto* mode      = &uart0_control::mode::asynchronous;
        constexpr static const auto* cs        = &character_size::cs8;
        constexpr static const auto  recv      = true;
        constexpr static const auto  transmit  = true;
    };
};

//------------------------------------------------------------------------------
// TWI
//------------------------------------------------------------------------------
struct twi_frequency {
    using freq_t  = uint32_t;
    using value_t = mc::word_t;

    constexpr static value_t calc_rate(freq_t f) {
        return ((cpu_frequency / f) - 16) / 2;
    }

    constexpr static value_t min_rate = 11;

    using set_rate = masks::write_reg<reg::twbr>;
};

struct twi_control {
    using M = masks;

    using enable            = M::mode_1<reg::twcr, TWEN>;
    using start             = M::P<reg::twcr, TWEN, TWINT, TWSTA>;
    using stop              = M::P<reg::twcr, TWEN, TWINT, TWSTO>;
    using step              = M::P<reg::twcr, TWEN, TWINT       >;
    using ack               = M::P<reg::twcr, TWEN, TWINT, TWEA >;
    using nack              = M::P<reg::twcr, TWEN, TWINT       >;
    using slave_ready       = M::P<reg::twcr, TWEN,        TWEA >;
    using transmit_complete = M::get<reg::twcr, TWINT>;
    using stop_pending      = M::get<reg::twcr, TWSTO>;
};

struct twi_mode {
    DEFINE_VALUE_ENUM(id_t, count_t)
    constexpr static id_t write = 0;
    constexpr static id_t read  = 1;
};

struct twi : public basic_twi <
    twi_frequency,
    twi_control,
    masks::read_word<reg::twsr, 0xF8>,
    masks::rw_reg<reg::twdr>,
    masks::write_reg<reg::twar>,
    twi_mode
> {};

//------------------------------------------------------------------------------
// SPI
//------------------------------------------------------------------------------
struct spi_control {
    struct mode {
        DEFINE_VALUE_ENUM(id_t, word_t);
        constexpr static id_t slave  = 0;
        constexpr static id_t master = 1;
        using set = masks::mode_1<reg::spcr, MSTR>;
    };

    struct data_order {
        DEFINE_VALUE_ENUM(id_t, word_t);
        constexpr static id_t msb = 0;
        constexpr static id_t lsb = 1;
        using set = masks::mode_1<reg::spcr, DORD>;
    };

    struct clock_polarity {
        DEFINE_VALUE_ENUM(id_t, word_t);
        constexpr static id_t lead_rising  = 0;
        constexpr static id_t lead_falling = 1;
        using set = masks::mode_1<reg::spcr, CPOL>;
    };

    struct clock_phase {
        DEFINE_VALUE_ENUM(id_t, word_t);
        constexpr static id_t lead_sample = 0;
        constexpr static id_t lead_setup  = 1;
        using set = masks::mode_1<reg::spcr, CPHA>;
    };

    struct clock_prescaler {
        DEFINE_VALUE_ENUM(id_t, word_t);
        constexpr static id_t ps4   = 0;
        constexpr static id_t ps16  = 1;
        constexpr static id_t ps64  = 2;
        constexpr static id_t ps128 = 3;
        constexpr static id_t ps2   = 4;
        constexpr static id_t ps8   = 5;
        constexpr static id_t ps32  = 6;
        using set = masks::mode_2_1<reg::spsr, SPI2X, reg::spcr, SPR1, SPR0>;
    };

    using enable = masks::mode_1<reg::spcr, SPE>;
};

struct spi_status {
    using transfer_complete = masks::get<reg::spsr, SPIF>;
    using write_collision   = masks::get<reg::spsr, WCOL>;
};

struct spi_pins {
    constexpr static const auto* mosi = &pin::spi::mosi;
    constexpr static const auto* miso = &pin::spi::miso;
    constexpr static const auto* sck  = &pin::spi::sck;
};

struct spi : public basic_spi <
    spi_control,
    spi_status,
    masks::rw_reg<reg::spdr>,
    spi_pins
> {
    struct defaults {
        constexpr static const auto* order
            = &spi_control::data_order::msb;
        constexpr static const auto* prescaler
            = &spi_control::clock_prescaler::ps64;
        constexpr static const auto* polarity
            = &spi_control::clock_polarity::lead_rising;
        constexpr static const auto* phase
            = &spi_control::clock_phase::lead_sample;
    };
};

//------------------------------------------------------------------------------
// DEVICE DEFINITION
//------------------------------------------------------------------------------
struct atmega328p {
    using reg    = mc::detail::reg;
    using pin    = mc::detail::pin;
    using irq    = mc::detail::irq;
    using adc    = mc::detail::adc;

    using timer0 = mc::detail::timer0;
    using timer1 = mc::detail::timer1;
    using timer2 = mc::detail::timer2;

    using uart   = mc::detail::uart0;
    using twi    = mc::detail::twi;
    using spi    = mc::detail::spi;

    using mask   = mc::detail::masks;

    struct defaults {
        using uart = mc::detail::uart0;
        using twi  = mc::detail::twi;
        using spi  = mc::detail::spi;
    };
};

using device = atmega328p;

//------------------------------------------------------------------------------
}} // namespace mc::detail

//------------------------------------------------------------------------------
#include "../common/footer.inc.hpp"
