#pragma once

//------------------------------------------------------------------------------
namespace mc { namespace io {

//------------------------------------------------------------------------------
template<class CharInput>
struct istream {
    istream(CharInput& dev): dev_(dev) {}
    CharInput& dev_;

    inline istream& operator>>(char& c) {
        c = dev_.getchar();
        return *this;
    }
};

//------------------------------------------------------------------------------
namespace detail {

//------------------------------------------------------------------------------
template<class Num,
         uint8_t Base,
         class Digit>
char* serialize_number(char* buf, Num n) {
    const bool neg = n < 0;
    char*      pos = buf;
    
    if (neg) n = -n;

    do {
        *pos++  = Digit::get(n % Base);
        n      /= Base;
    } while(n > 0);

    if (neg) *pos++ = '-';

    return pos;
}

//------------------------------------------------------------------------------
template<class   OStream,
         class   Num,
         uint8_t Base,
         class   Digit,
         uint8_t BufferSize>
void print_number(OStream& os, Num n) {
    char buf[BufferSize];
    auto* p = serialize_number<Num, Base, Digit>(buf, n);
    while(p != &buf[0]) os << *--p;
}

//------------------------------------------------------------------------------
struct decimal_digit {
    template<class Num>
    static inline Num get(Num n) { return '0' + n; }
};

template<class OStream, class Num>
inline void print_decimal_number(OStream& os, Num n) {
    print_number<OStream, Num, 10, decimal_digit, 13>(os, n);
}

//------------------------------------------------------------------------------
struct hex_digit {
    template<class Num>
    static inline Num get(Num n) {
        return (n <= 9 ? '0'       
                       : ('A' - 10))
               + n;
    }
};

template<class OStream, class Num>
inline void print_hex_number(OStream& os, Num n) {
    print_number<OStream, Num, 16, hex_digit, 13>(os, n);
}

template<class Num>
struct as_hex {
    as_hex(Num n) : n_(n) {}
    Num operator*() const { return n_; }
    Num n_;
};

//------------------------------------------------------------------------------
} // namespace detail

template<class Num>
inline detail::as_hex<Num> hex(Num n) { return {n}; }

//------------------------------------------------------------------------------
template<class CharOutput>
struct ostream {
    ostream(CharOutput& dev): dev_(dev) {}
    CharOutput& dev_;

    inline ostream& operator<<(char c) {
        dev_.putchar(c);
        return *this;
    }

    inline ostream& operator<<(const char* c) {
        for(; *c != 0; ++c) {
            *this << *c;
        }

        return *this;
    }

//------------------------------------------------------------------------------
    inline ostream& operator<<(detail::as_hex<uint8_t> h) {
        detail::print_hex_number<ostream, uint8_t>(*this, *h); return *this;
    }

    inline ostream& operator<<(detail::as_hex<int8_t> h) {
        detail::print_hex_number<ostream, int8_t>(*this, *h); return *this;
    }

    inline ostream& operator<<(detail::as_hex<uint16_t> h) {
        detail::print_hex_number<ostream, uint16_t>(*this, *h); return *this;
    }

    inline ostream& operator<<(detail::as_hex<int16_t> h) {
        detail::print_hex_number<ostream, int16_t>(*this, *h); return *this;
    }

    inline ostream& operator<<(detail::as_hex<uint32_t> h) {
        detail::print_hex_number<ostream, uint32_t>(*this, *h); return *this;
    }

    inline ostream& operator<<(detail::as_hex<int32_t> h) {
        detail::print_hex_number<ostream, int32_t>(*this, *h); return *this;
    }

//------------------------------------------------------------------------------
    inline ostream& operator<<(uint16_t n) {
        detail::print_decimal_number<ostream, uint16_t>(*this, n); return *this;
    }

    inline ostream& operator<<(int16_t n) {
        detail::print_decimal_number<ostream, int16_t>(*this, n); return *this;
    }

    inline ostream& operator<<(uint32_t n) {
        detail::print_decimal_number<ostream, uint32_t>(*this, n); return *this;
    }

    inline ostream& operator<<(int32_t n) {
        detail::print_decimal_number<ostream, uint32_t>(*this, n); return *this;
    }
};

//------------------------------------------------------------------------------
}} // namespace mc::print
