#pragma once

namespace mc { namespace util {
//------------------------------------------------------------------------------

// LEVEL GUARD
template<const pin::id_t& Pin>
struct level_guard {
    level_guard(pin::level hold_level) {
        old_level = pin::get<Pin>();
        pin::set<Pin>(hold_level);
    }

    ~level_guard() {
        pin::set<Pin>(old_level);
    }

    pin::level old_level;
};

//------------------------------------------------------------------------------
// PULSE / FLASH / BLINK / ...
template<const pin::id_t& Pin, pin::level Level = HIGH>
void pulse() {
    pin::set<Pin>(Level);
    asm("nop");
    pin::set<Pin>(pin::inverse_level(Level));
}

template<const pin::id_t& Pin, pin::level Level = HIGH>
void blink(count_t times = 5, count_t delay = 50) {
    using namespace mc;

    for (count_t i = 0; i < times; ++i) {
        pin::set<Pin>(Level);
        util::wait(delay);
        pin::set<Pin>(pin::inverse_level(Level));
        util::wait(delay);
    }
}

//------------------------------------------------------------------------------
}} // namespace mc::util
