#pragma once

//------------------------------------------------------------------------------
namespace mc { namespace util { namespace spi {

//------------------------------------------------------------------------------
namespace detail {

//------------------------------------------------------------------------------
template<class SPI>
struct transceiver {
    inline word_t transceive(word_t w) {
        SPI::write(w);
        while(! SPI::transfer_complete());
        return SPI::read();
    }

    inline word_t operator()(word_t w) {
        return this->transceive(w);
    }
};

//------------------------------------------------------------------------------
} // namespace detail

//------------------------------------------------------------------------------
template<class SPI>
struct master : public detail::transceiver<SPI> {
    master() { this->reset(); }

    inline void reset() {
        SPI::template set_mode<SPI::mode::master>();
        device::pin::configure<*SPI::pin::mosi>(OUTPUT);
        device::pin::configure<*SPI::pin::miso>(INPUT);
        device::pin::configure<*SPI::pin::sck >(OUTPUT);
    }
};

//------------------------------------------------------------------------------
template<class SPI>
struct slave : public detail::transceiver<SPI> {
    slave() { this->reset(); }

    inline void reset() {
        SPI::template set_mode<SPI::mode::slave>();
        device::pin::configure<*SPI::pin::mosi>(INPUT);
        device::pin::configure<*SPI::pin::miso>(OUTPUT);
        device::pin::configure<*SPI::pin::sck >(INPUT);
    }
};

//------------------------------------------------------------------------------
template<
    class SPI,
    const typename SPI::prescaler::id_t& Prescaler = *SPI::defaults::prescaler,
    const typename SPI::polarity::id_t&  Polarity  = *SPI::defaults::polarity,
    const typename SPI::phase::id_t&     Phase     = *SPI::defaults::phase,
    const typename SPI::order::id_t&     Order     = *SPI::defaults::order
>
struct context {
    context() {
        this->reset();
    };

    ~context() {
        SPI::disable();
    };

    void reset() {
        SPI::enable();
        SPI::template set_polarity <Polarity >();
        SPI::template set_phase    <Phase    >();
        SPI::template set_prescaler<Prescaler>();
        SPI::template set_order    <Order    >();
    }

    mc::util::spi::master<SPI> master() { return {}; }
    mc::util::spi::slave <SPI> slave () { return {}; }
};

//------------------------------------------------------------------------------
}}} // namespace mc::util::spi
